%{
  Format: [offloading_ratio,x]=greedy_codedcaching(Nu,Nf,pf,C,B,K,T,lambda);
  Input: Nu: an integer denoting the number of users
         Nf: an integer denoting the number of files
         pf: 1*Nf vector denoting the file request probability, where pf(f) is corresponding to file f
         C: a number denoting the caching capacity
         B: an integer denoting the number of segment can be transmitted in one contact time
         K: 1*Nf vector where K(f) is the number of segments to recover file f
         T: an integer denoting the tolerate delay time
         lambda: Nu*Nu matrix, where lambda(i,j) denotes the contact rate of users i and j
  Output: offloading_ratio: a number denoting the maximum data offloading ratio
          x: Nu*Nf matrix denoting the optimal caching placement, where x(i,f) denotes the number of segments of file f cached at user i
  Remark: Before using this function, The C++ codes, i.e., pdf_fast.cpp, should be compiled by 'mex' in matlab. Please refer to http://www.mathworks.com/help/matlab/ref/mex.html 
%}

function [offloading_ratio,x]=greedy_codedcache_fast(Nu,Nf,pf,C,B,K,T,lambda)

x=zeros(Nu,Nf);
count=zeros(1,Nu);
add_item=zeros(Nu,Nu);
index=ones(Nu,Nf);
add_max=zeros(Nf,2);  % maximum priority value of each file
add_sum=zeros(Nu,Nf); % priority value
offloading_ratio=0;
jx=ones(max(K)+1,1);
for (i=3:max(K)+1) jx(i)=jx(i-1)*(i-1);
end

% initialize the priority values
for (j=1:Nu)
    for (i=1:Nu)
        add_item(j,i)=1/Nu*(1-exp(-T*lambda(i,j)));
    end
end  
for (k=1:Nf)
    add_sum(:,k)=pf (k)/K(k)*add_item(:,:)*index(:,k);
    [add_max(k,1),add_max(k,2)]=max(add_sum(:,k));
end

% greedy algorithm
for (j=1:Nu*C)
    
    % select the element with maximum priority value
    [~,kx]=max(add_max(:,1));
    yx=add_max(kx,2);
    offloading_ratio=offloading_ratio+add_max(kx,1);
    
    % update the remaining set, which includes the remaining elements that can be added considering the constraint
    count(yx)=count(yx)+1;
    if (count(yx)>=C)
        add_sum(yx,:)=ones(1,Nf)*(-inf);
        for (k=1:Nf)
            [add_max(k,1),add_max(k,2)]=max(add_sum(:,k));
        end
    end    
    x(yx,kx)=x(yx,kx)+1;
    if (x(yx,kx)>=K(kx))
        add_sum(yx,kx)=-inf;
    end
    
    % update the priority values
	for (j1=1:Nu)
		if (add_sum(j1,kx)>=0) add_sum(j1,kx)=0;
        end
    end  
    for (i=1:Nu)
		if (K(kx)-x(i,kx)-1>=0)
			temp_prob=pdf_fast(Nu,K(kx)-x(i,kx)-1,lambda,x(:,kx),B,T,i-1,i-1);
            if (add_sum(i,kx)>=0) add_sum(i,kx)=add_sum(i,kx)+temp_prob(K(kx)-x(i,kx))*pf(kx)/Nu/K(kx);
            end
            for (j1=1:Nu)
                if (i ~= j1 && add_sum(j1,kx)>=0)
                    if (x(j1,kx)==0) add_sum(j1,kx)=add_sum(j1,kx)+gammainc(T*lambda(i,j1),ceil((x(j1,kx)+1)/B))*temp_prob(K(kx)-x(i,kx))*pf(kx)/Nu/K(kx);
                    else if (K(kx)-x(i,kx)-x(j1,kx)-1>=0)
                            temp_k=K(kx)-x(i,kx)-x(j1,kx);
                            temp_a=zeros(temp_k,temp_k+1);
                            prob_j0=exp(-T*lambda(i,j1));
                            for (li=1:temp_k) 
                                temp_a(li,li)=1;
                                temp_a(li,temp_k+1)=temp_prob(temp_k-li+1)/prob_j0;
                            end

                            prob_j=zeros(temp_k-1,1);
                            for (z=1:min(x(j1,kx)-1,temp_k-1))
                                if (mod(z,B) == 0)
                                    prob_j(z)=(T*lambda(i,j1))^(z/B)/(jx(z/B+1));
                                else
                                    prob_j(z)=0;
                                end
                            end
                            if (x(j1,kx)<=temp_k-1) prob_j(x(j1,kx))=(1-gammainc(floor(x(j1,kx)/B),T*lambda(i,j1)))/prob_j0;
                            end
                            for (li=1:temp_k-1)
                                for (lj=li+1:temp_k)
                                    temp_a(li,lj)=prob_j(lj-li);
                                end
                            end
                            for (li=2:temp_k) temp_a(1,:)=temp_a(1,:)-temp_a(1,li)*temp_a(li,:);
                            end
                            add_sum(j1,kx)=add_sum(j1,kx)+gammainc(T*lambda(i,j1),ceil((x(j1,kx)+1)/B))*temp_a(1,temp_k+1)*pf(kx)/Nu/K(kx);
                        end
                    end
                end
            end
        end
    end
    [add_max(kx,1),add_max(kx,2)]=max(add_sum(:,kx));
end
