/*
  Format: p=pdf_fast(Nu,K,lambda,x(:,f),B,T,i1,i2)
  Input: Nu: number of users
         K: a non negative integer
         lambda: pairwise contact rate
         x(:,f): caching placement of file f
         B: number of segment can be transmitted in one contact time
         T: tolerate delay time
         i1,i2: indexes of users
  Output: p is a vector with (K+1) elements, where 
          in C++: p(i)=Pr[\sum_{j \ne i1,i2} \min ( B M_{i1,j} , x_{j,f} ) <= K-i] 
          in Matlab: p(i)=Pr[\sum_{j \ne i1,i2} \min ( B M_{i1,j} , x_{j,f} ) <= K-i+1] 
 
  Remark: This function needs to use 'mex' to run in Matlab. 
          Please refer to http://www.mathworks.com/help/matlab/ref/mex.html
*/

#include <iostream>
#include <vector>
#include <fstream>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "mex.h"
#define OUTM plhs[0]
#define MAXN 100   //should be larger than number of users. Please adjust this constant value according to the number of users.
#define MAXK 100 //should be larger than K. Please adjust this constant value according to K.

using namespace std;


double prob(int j,double lambda[MAXN][MAXN],int x[MAXN],int B,int T,int r,int *jx,int nu, int i1, int i2, int visit[MAXN][MAXN], double pr[MAXN][MAXN])
{
    double result=0,temp;
    while (j==i1 || j==i2) j=j+1;
    if (visit[j][r])
        return pr[j][r];
    else
    {
        if (j<nu-1)
        {
            temp=0;
            double lamPow=exp(-lambda[i1][j]*T);
            for (int z=0;z*B<x[j] && B*z<=r; z++)
            {
                double tt=lamPow/jx[z];
                temp=temp+tt;
                result=result+tt*prob(j+1,lambda,x,B,T,r-z*B,jx,nu,i1,i2,visit,pr);
                lamPow*=lambda[i1][j]*T;
            }
            if (x[j]<=r)
                result=result+(1-temp)*prob(j+1,lambda,x,B,T,r-x[j],jx,nu,i1,i2,visit,pr);
        }
        else
        {
            if (x[j]<=r)
                result=1;
            else
            {
                double lamPow=exp(-lambda[i1][j]*T);
                for (int z=0;B*z<=r; z++)
                {
                    result=result+lamPow/jx[z];
                    lamPow*=lambda[i1][j]*T;
                }
            }
        }
        visit[j][r]=1;
        pr[j][r]=result;
        return result;
    }
}


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    double *data=NULL,*output=NULL,lambda[MAXN][MAXN],pr[MAXN][MAXK];
    int nu,nf,C,T,B,K,i1,i2,x[MAXN],visit[MAXN][MAXK],jx[MAXK];
    //visit is used to avoid repetitive computation
    //pr[j][r]=Pr[sum from the j-th term to the last term <= r]    
    
    data=mxGetPr(prhs[0]);
    nu=data[0];
    data=mxGetPr(prhs[1]);
    K=data[0];
    data=mxGetPr(prhs[2]);
    if (K>=0) jx[0]=1;
    if (K>=1) jx[1]=1;
    for (int i=2; i<K; i++) jx[i]=jx[i-1]*i;
    
    for (int i=0;i<nu;i++)
        for (int j=0;j<nu;j++)
        {
            lambda[j][i]=data[i*(nu)+j];
        }
    data=mxGetPr(prhs[3]);
    for (int i=0;i<nu;i++)
        x[i]=data[i];
    data=mxGetPr(prhs[4]);
    B=data[0];
    data=mxGetPr(prhs[5]);
    T=data[0];
    data=mxGetPr(prhs[6]);
    i1=data[0];
    data=mxGetPr(prhs[7]);
    i2=data[0];

    for (int i=0;i<nu;i++)
        for (int j=0;j<=K;j++)
        {
            visit[i][j]=0;
            pr[i][j]=0;
        }

    OUTM= mxCreateDoubleMatrix(1,K+1, mxREAL);
    output=mxGetPr(OUTM);

    while (nu-1==i1 || nu-1==i2) nu=nu-1;
    output[K]=prob(0,lambda,x,B,T,K,jx,nu,i1,i2,visit,pr);
    for (int i=0; i<K; i++) output[i]=pr[0][i];
}
